;       Program:        MusClick.asm

;       Programmers:    Jay Parsons (Jparsons)
;                       and Ken Chan (HazMatZak)
;       Inspiration:    Bowen Moursund (Bowen)
;                       and the folks at BORBBS
;       Date:           July 6, 1992
;       Version:        1.1

;       Assembly syntax:

;               MASM musclick    or     TASM musclick
;               LINK musclick           TLINK musclick
;               EXE2BIN musclick        EXE2BIN musclick

;       dBASE usage syntax, after LOAD musclick:

;               cSuccess = call( "MUSCLICK", "I", [<arg2>] ) && install
;               cSuccess = call( "MUSCLICK", "U" )           && uninstall

;               store chr( 255 ) TO cMRow, cMCol
;               call MUSCLICK with cMRow, cMCol              && report row, col
;               nMRow = asc( cMRow )
;               nMCol = asc( cMCol )

;       A call with "I" must be made to install the alternate mouse
;       event handler for this to work.  After that call, this .bin
;       will push a Ctrl-PgUp, or the character specified in the optional
;       <arg2> argument to the "I" call, into the BIOS buffer when
;       the left mouse button is released.  The <arg2> character must be
;       a chr() value from 1 to 254, or the argument may be a string
;       composed of two chr() values, the first of which is chr(255) and
;       the second of which is the chr() value of the second (high) byte
;       of an extended code for a key or key combination that has no
;       ASCII value.

;       For example, a call with a second argument of chr(255) + chr(97)
;       will stuff Ctrl-F4.  It is likely that you will want to use a
;       key or key combination that has an inkey() and that works with
;       ON KEY.  Note that the extended codes are always positive
;       and not always the same as the inkey() values.

;       The installation call will return "T" if successful, and "F" if not.

;       If the BIOS buffer happens to be full when the mouse button is
;       released, the click will go unreported, but the row and column
;       will be updated.

;       If you wish a default character different from Ctrl-PgUp, change
;       the DEFAULTKEY value before assembly.  For an extended code,
;       set it to the value of the second byte * 256 (100h).  The ESC key
;       is the only key that will attract dBASE's attention in the middle
;       of a dBASE operation such as "LIST" or "INDEX".  However, that
;       operation will be cancelled if ESC is pressed, so consider what is
;       going on and whether you want to have it recognize a click.

;       Calls to get the row and column must have the first argument
;       be chr(255), and must have a second argument.  Both MUST be variables.
;       A call to find the row and column if no "I" call has been made,
;       or no clicks have occurred, will return chr(255) for each.
;       The row and column reported are those as of the last click, which are
;       not necessarily those current at the time of the call.  They are
;       determined by dividing the values in pixels reported by the mouse
;       by 8 each and ignoring any remainder.

;       After one call has been made to get the mouse row and column,
;       subsequent calls to get the row and column will return chr(254) for
;       each if no click has occurred since the previous call.  This allows
;       the dBASE program to differentiate between actual keypresses
;       and "keypresses" made by this routine to report mouse clicks.

;       The recommended way to test for an actual mouse click is to see if the
;       row value returned is below 200; all row values 200 and above are
;       reserved for future expansion (only 254 and 255 are currently used).

;       A call with "U" should be made to uninstall the alternate event
;       handler before installing it with a different key, and also before
;       ending the dBASE program for safety.  Repeated calls to install
;       the handler without uninstalling it will overwrite the address of
;       the dBASE handler and prevent restoring it.

;       The dBASE mouse-event handler is disabled while the one in this
;       .bin is installed.  Consequently, dBASE will not know of the
;       current mouse position or click status, and calls to the .bins
;       in the recent Technotes article, uploaded as MOUSE_FU.TXT, will
;       not return correct information.  Use this .bin or those, but not
;       both at once.

;               Following are as specified by Microsoft for its mouse:

;       Register usage on call of an alternate event handler

;               AX = condition mask ( bits 0-4 set for events occurred )
;               BX = button state   ( bit 0 set if left button pressed;
;                                     bit 1 if right; bit 2, center.) <unused>
;               CX = horizontal mouse cursor coordinate in pixels
;               DX = vertical    "     "        "        "   "
;               SI = last horizontal mickey count ( vertical? )   <unused>
;               DI = last vertical mickey count   ( horizontal? ) <unused>
;               DS = mouse driver data segment <unused>

;       Event mask bits, 1 to respond to that event:
;                0 = mouse has moved       ( value  1 )
;                1 = left button pressed   ( value  2 )
;                2 = left button released  ( value  4 )
;                3 = right button pressed  ( value  8 )
;                4 = right button released ( value 16 )

;       Thus, if our handler were to be called whenever the mouse moved
;       or either button were released, the mask would be 10101, or
;       16 + 4 + 1, decimal value 21, rather than the 4 we use because
;       we care only about release of the left button.

;       If an event not specified in our event mask occurs, our handler
;       will not be called at all.  If the left button is released, it
;       will be called.  AX, the condition mask, will report all events
;       that have occurred, not only the one specified in the event mask.

;               program variables

DEFAULTKEY      equ     31              ; default character simulated,
                                        ; US (ASCII Unit Separator), which
                                        ; dBASE maps to Ctrl-PgUp
RELEASELEFT     equ      4              ; release of left key, bit 2
REALKEY         equ     254*257         ; 0FEFEh to return for real keypress

;               BIOS keyboard data address components

BIOSDATA        equ     40h             ; BIOS data segment
BUFFHEAD        equ     001Ah           ; offset of buffer head pointer
BUFFTAIL        equ     001Ch           ; offset of buffer tail pointer
KBBUFF          equ     001Eh           ; buffer itself, 32 bytes
KBBUFFEND       equ     KBBUFF + 32

;               dBASE key flag value

DBKEY           equ     1               ; 1 signals a key has been hit.

;               80x8x opcodes to look for to find flag address

MOVWORD         equ     0C7h            ; opcode for move immediate word
IMMED           equ     6               ; opcode for immediate move address

                .MODEL tiny
                .CODE

                jmp short start

keypress        dw DEFAULTKEY           ; default "ON KEY" label
oldoff          dw ?                    ; address of previous mouse handler
oldseg          dw ?
oldmask         dw ?                    ; and its event mask
row             db 255                  ; row of mouse cursor
column          db 255                  ; column ditto

dbkeyhitflag    dw      0               ; offset address of dBASE key-hit flag
dbdataseg       dw      ?               ; and its segment address

start:          mov bp,cs               ; save code segment address
                cmp cx,1
                jb allret               ; quit if no arguments
                lds si,es:[di]          ; point to first argument
                lodsb                   ; fetch it
                mov ds,bp               ; data is with our code
                cmp al,255              ; is first argument chr(255)?
                jz report               ; then jump to report row and column
                and al,5Fh              ; capitalize the argument
                cmp al,'I'
                jz install              ; jump to install event handler
                cmp al,'U'
                jz uninstall            ; or to uninstall it

badret:         mov al,'F'              ; if none of the above, return 'F'
                jmp short argret

report:         mov ax,word ptr row     ; get column to ah, row to al
                mov word ptr row,REALKEY; set row and column to "real key"
                                        ; settings
                lds si,es:[di+4]        ; point to 2d argument
                mov [si],ah             ; save column there

argret:         les di,es:[di]          ; point to 1st argument
                stosb                   ; save row or T/F value
allret:         retf                    ; return to dBASE

;       This version of MUSCLICK scans through dBASE's Int 09h handler,
;       trying to determine the location of dBASE's "key has been hit"
;       flag by looking for a MOV WORD PTR in handler, which has so far
;       been part of dBASE's own instruction to set that flag.

install:        mov bx,sp
                mov ax,ss:[bx+4]        ; get word dBASE pushed onto stack
                mov dbdataseg,ax        ; save the dBASE data segment
                push cx                 ; save argument count
                push es                 ; and argument table pointers
                push di
                xor ax,ax
                mov es,ax               ; segment absolute 0
                les di,es:[9*4]         ; get int 9 vector to es:di
                mov cx,60               ; look thru 60 bytes of int 9 handler
                mov al,MOVWORD          ; for "mov word ptr" opcode
findloop:       repne scasb             ; until we find a match
                jcxz notfound           ; if not found, we are lost
                cmp byte ptr es:[di],IMMED ; is "direct immediate" opcode next?
                jnz findloop            ; if it's not, try again
                mov ax,word ptr es:[di+1] ; success! get the flag address
                mov dbkeyhitflag,ax     ; and save it
notfound:       pop di
                pop es
                pop cx
                cmp dbkeyhitflag,0      ; check for flag address found
                jz badret               ; if not found, return false

twoargs:        cmp cx,2
                jb onearg
                lds si,es:[di+4]        ; get second argument
                lodsw                   ; to ax
                cmp al,255              ; is it extended scan code?
                jz scancode             ; then jump
                xor ah,ah               ; blank second byte if not already
                jmp short savekey
scancode:       xor al,al               ; or blank the first byte
savekey:        mov ds,bp               ; and restore segment
                mov keypress,ax

onearg:         mov bx,bp
                mov dx,offset clickhandler
                mov cx,RELEASELEFT      ; event mask for left button release
                jmp short swap

uninstall:      mov bx,oldseg           ; restore these
                mov dx,oldoff
                mov cx,oldmask

swap:           push es
                mov es,bx
                mov ax,14h              ; swap handlers
                int 33h                 ; call the mouse driver
                mov bx,es               ; save segment of old handler
                pop es

saveold:        mov oldoff,dx           ; save old address
                mov oldseg,bx
                mov oldmask,cx          ; and old mask

                mov al,'T'              ; and return a 'T'
                jmp argret

;--------------------------------------------------------------------
;               Event handler for a click, release of left button

;       see above for state of registers on call to this handler by the
;       mouse driver.

;       This handler converts the vertical and horizontal pixel counts
;       to row and column, saves those in the .bin's memory, pushes
;       the designated key into the BIOS buffer unless it is full, and
;       sets the dBASE "a-key-is-hit" flag so it will trigger an ON KEY.

;       This code cannot use interrupts due to lack of reentrancy of DOS
;       and the BIOS.  This code is largely adapted from an old BIOS listing
;       for an int 9 handler.

clickhandler:   push ds                 ; save mouse driver segment
                mov ax,cs               ; point ds to our code
                mov ds,ax
                mov ax,cx               ; save the horizontal pixels for now
                mov cx,3                ; shift count is 3 bits, divide by 8
                shr dx,cl               ; divide vertical pixels for row
                mov row,dl              ; save the row
                shr ax,cl               ; divide horizontal pixels for column
                mov column,al           ; and save the column
                mov ax,BIOSDATA         ; point to BIOS data segment
                mov ds,ax
                mov ax,ds:[BUFFTAIL]    ; get pointer to tail of buffer
                mov si,ax               ; save it here for now
                add ax,2                ; advance the tail
                cmp ax,KBBUFFEND        ; time to wrap around?
                jnz nowrap              ; no
                mov ax,KBBUFF           ; yes, point to start
nowrap:         cmp ax,ds:[BUFFHEAD]    ; is tail now up to the head?
                jz fullbuff             ; if so, do nothing, buffer is full
                mov bx,word ptr cs:keypress
                mov [si],bx             ; push two-byte scan code
                mov ds:[BUFFTAIL],ax    ; and adjust the tail
                lds si,dword ptr cs:[dbkeyhitflag] ; get dBASE flag address
                mov word ptr ds:[si],DBKEY ; set the flag
fullbuff:       pop ds                  ; restore mouse driver segment
clickret:       retf                    ; and return to mouse driver

                end
